----------------------------------------------------------------------------------
-- Engineer:       Stanislav Maslan, s.maslan@seznam.cz
-- www:            http://www.elektronika.kvalitne.cz
-- www:            http://www.elektronika.kvalitne.cz/PLA/TFT_serial/TFT_serial.html
--
-- Create Date:    2012.05.16
-- Target Devices: CoolRunner2 XC2C256-7
-- Project Name:   640x480x16-bit TFT controller (UART interface)

-- Description:
-- ------------
-- Simple serial controller for TFT display. Received data is stored into ext.
-- SRAM (512k x 16, 10ns) which is shared for TFT refresh. Maximal serial rate 
-- is some 12Mbit/s for 48MHz master clock. Receive timeout is some 500ms. After
-- this interval the write address is reseted to 0.
----------------------------------------------------------------------------------
---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;


-- libs
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity main is
	------ ports ------
	port(
		-- source clock (48MHz)
		iclk: in std_logic;
		-- TFT pixel clock
		pclk: out std_logic;
		-- TFT synchronizations
		hsync: out std_logic;
		vsync: out std_logic;
		-- TFT color data
		red: out std_logic_vector(4 downto 0);
		grn: out std_logic_vector(5 downto 0);
		blu: out std_logic_vector(4 downto 0);
		-- UART recieve line
		rxd: in std_logic;
		-- status LED
		LED: out std_logic;
		-- SRAM data bus
		D: inout std_logic_vector(15 downto 0);
		-- SRAM address bus
		A: out std_logic_vector(18 downto 0);
		-- SRAM write
		WR: out std_logic;
		-- SRAM output enable
		RD: out std_logic);

	------ TFT definition constants ------
	-- TFT clocking parameters
	constant xw: integer := 790;
	constant yw: integer := 520;
	constant hsw: integer := 16;
	constant vsw: integer := 16;
	
	-- TFT visible area
	constant xofs: integer := 100;
	constant yofs: integer := 34;
	constant xsize: integer := 640;
	constant ysize: integer := 480;
	
	-- TFT color ranges
	constant rmax: integer := 31;
	constant gmax: integer := 63;
	constant bmax: integer := 31;
	
	------ UART setup ------
	-- clock divisor: baud_rate = XTAL/(2*ubrr)
	constant ubrr: integer := 8;
	-- write address reset timeout (vsync cycles, choose 2^M values)
	constant urst: integer := 32;
end main;


architecture Behavioral of main is
	-- internal clock
	signal xclk: bit := '0';
	signal yclk: bit := '0';
	-- TFT position coutners
	signal pcnt: integer range 0 to xw;
	signal rcnt: integer range 0 to yw;
	
	-- SRAM address registers
	signal rda: integer range 0 to 2**19-1;
	signal wra: integer range 0 to 2**19-1;
	-- SRAM write address reset timout
	signal wrst: integer range 0 to urst-1;
	
	-- UART clock divisor
	signal udiv: integer range 0 to ubrr-1 := 0;
	-- UART bit counter
	signal ucnt: integer range 0 to 15 := 0;
	-- UART byte shift register
	signal udat: std_logic_vector(6 downto 0);
	-- UART word buffer
	signal umem: std_logic_vector(15 downto 0);
	-- UART recieve complete pulse flag
	signal rxcf: std_logic := '0';
	-- UART low/high byte flag
	signal rxhi: std_logic := '0';
begin
	
	process(iclk,xclk,pcnt,rcnt)
		variable frame: boolean;
		variable fsync: boolean;		
	begin
		
		------ master clock divider ------
		if(iclk'event) then
			--- internal clocks ---
			-- 0deg clock
			xclk <= yclk;
			-- 90deg clock
			yclk <= not xclk;
		end if;
		
		
		
		
		------ sync counters ------
		if(xclk='0' and xclk'event) then
			-- pixel counter
			if(pcnt>=xw) then
				pcnt <= 0;
				
				-- row counter
				if(rcnt>=yw) then
					rcnt <= 0;
				else
					rcnt <= rcnt + 1;
				end if;
			else
				pcnt <= pcnt + 1;
			end if;
		end if;
		
		-- internal frame sync signal
		fsync := pcnt=0 and rcnt=0;
		
		-- internal valid area signal
		frame := pcnt>=xofs and pcnt<(xofs+xsize) and rcnt>=yofs and rcnt<(yofs+ysize);
		
		
		
		
		------ TFT refresh ------
		-- SRAM read address counter
		if(fsync) then
			rda <= 0;
		elsif(frame and xclk'event and xclk='0') then
			rda <= rda + 1;
		end if;
		
		-- TFT color data <= SRAM data word
		if(xclk='0') then
			red <= D(4 downto 0);
			grn <= D(10 downto 5);
			blu <= D(15 downto 11);
		end if;
		
		
		
		
		------ SRAM bus multiplexer ------
		if(xclk='0') then
			-- read cycle
			A <= conv_std_logic_vector(rda,A'length);
			D <= "ZZZZZZZZZZZZZZZZ";
		else
			-- write cycle
			A <= conv_std_logic_vector(wra,A'length);
			D(4 downto 0) <= umem(4 downto 0);
			D(10 downto 5) <= umem(10 downto 5);
			D(15 downto 11) <= umem(15 downto 11);
		end if;
				
		
		
		
		------ UART reciever ------
		if(xclk'event and xclk='0') then
			
			if(ucnt=0 and rxd='0') then
				--- start bit detected, delay to ubrr/2 ---
				udiv <= ubrr/2;
				ucnt <= ucnt + 1;
							
			elsif(udiv>=(ubrr-1)) then
				--- bit clock ---
				
				-- reset baud rate divider
				udiv <= 0;
				
				if(ucnt=9) then
					-- final bit
					if(rxhi='1') then
						umem(15 downto 8) <= rxd&udat(6 downto 0);
					else
						umem(7 downto 0) <= rxd&udat(6 downto 0);
					end if;
					
					-- low/high byte flag
					rxhi <= not rxhi;
				elsif(ucnt>1 and ucnt<9) then
					-- not final data bit
					udat(6 downto 0) <= rxd&udat(6 downto 1);
				end if;
				
				-- rx flag pulse
				if(ucnt=9 and rxhi='1') then
					-- rx complete
					rxcf <= '1';
				end if;
				
				-- increment bit counter
				if(ucnt>9) then
					-- stop
					ucnt <= 0;
				else
					-- next bit
					ucnt <= ucnt + 1;
				end if;
				
			elsif(ucnt>0) then
				-- clock divider
				udiv <= udiv + 1;
				rxcf <= '0';
				
			else
				-- clear rxc flag
				rxcf <= '0';
				
			end if;			
			
		end if;
		
		
		
		
		------ recieve timout counter ------
		if(rxcf='1') then
			wrst <= 0;
		elsif(xclk'event and xclk='0' and fsync) then
			if(wrst<(urst-1)) then
				wrst <= wrst + 1;
			else
				wrst <= 0;
			end if;
		end if;
		
		
		
		
		------ increment or reset write address ------
		if(wrst=(urst-1)) then
			wra <= 0;
		elsif(rxcf'event and rxcf='0') then
			wra <= wra + 1;
		end if;
		
	end process;
	
	------ status LED ------
	LED <= '0' when wra=640*480 else '1';
	
	
	------ SRAM write/read pulses ------
	RD <= '0' when xclk='0' and yclk='1' else '1';
	WR <= '0' when xclk='1' and yclk='0' and rxcf='1' else '1';
	
	
	------ TFT clocks generator ------
	-- pixel clock
	pclk <= '1' when xclk='1' and yclk='1' else '0';
	
	-- hsync
	hsync <= '1' when pcnt<(xw-hsw) else '0';
			
	-- vsync
	vsync <= '1' when rcnt>vsw else '0';
		
end Behavioral;

